# Competitive Policy Gradient (CoPG) Algorithm
This repository contains all code and experiments for competitive policy gradient (CoPG) algorithm. The paper for competitive policy gradient can be found [here],
The code for Trust Region Competitive Policy Optimization (TRCPO) algorithm can be found [here]

[square brackets show link] (links will be added later, to keep it anonymous)

## Experiment videos are available [here](https://sites.google.com/view/rl-copo)
## Dependencies
1. Code is tested on python 3.5.2.
2. Only Markov Soccer experiment requires [OpenSpiel library](https://github.com/deepmind/open_spiel), Other 5 experiments can be run directly. 
3. Require [torch.utils.tensorboard](https://pytorch.org/docs/stable/tensorboard.html)

## Repository structure
    .
    ├── notebooks
    │   ├── LQ_game.ipynb
    │   ├── bilinear_game.ipynb
    │   ├── RockPaperScissors.ipynb
    │   ├── matching_pennies.ipynb
    │   ├── MarkovSoccer.ipynb
    │   ├── CarRacing.ipynb
    ├── game                            # Each game have a saparate folder with this structure
    │   ├── game.py                     
    │   ├── copg_game.py                
    │   ├── gda_game.py
    │   ├── network.py
    │   ├── pretrained_models.py       (if applicable)
    │   ├── results.py                 (if applicable)
    ├── copg_optim
    │   ├── copg.py 
    │   ├── critic_functions.py 
    │   ├── utils.py 
    ├── car_racing_simulator
    └── ...
1. [Jupyter notebooks] are the best point to start. It contains demonstrations and results. 
2. Folder [copg_optim] contains optimization code

## How to start ?
Open jupyter notebook and run it to see results.

or

```
git clone "adress"
cd copg
cd RockPaperScissors
python3 copg_rps.py
cd ..
cd tensorboard
tensordboard --logdir .
```
You can check results in the tensorboard.

## Experiment Demonstration
### &nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp; GDA vs GDA   &nbsp;&nbsp;&nbsp;&nbsp;  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;  &nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;                  CoPG vs CoPG
### ORCA Car Racing
&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;![](https://user-images.githubusercontent.com/37489792/84293831-98562980-ab48-11ea-8a7a-1d65f40e5f15.gif) &nbsp; &nbsp;&nbsp;&nbsp;&nbsp; &nbsp; &nbsp;&nbsp;&nbsp;&nbsp; &nbsp; &nbsp;&nbsp;&nbsp;&nbsp; ![](https://user-images.githubusercontent.com/37489792/84293827-968c6600-ab48-11ea-9756-af3c4092175a.gif)
### Rock Paper Scissors
&nbsp; &nbsp; &nbsp;&nbsp;&nbsp; &nbsp;&nbsp;<img src="https://user-images.githubusercontent.com/37489792/84292894-495bc480-ab47-11ea-97ae-ff326d6d6c18.gif" width="350" height="250">&nbsp; &nbsp;&nbsp;<img src="https://user-images.githubusercontent.com/37489792/84292908-4cef4b80-ab47-11ea-8de5-897b1f9c7b52.gif" width="70" height="250">&nbsp; &nbsp;&nbsp;&nbsp; <img src="https://user-images.githubusercontent.com/37489792/84292897-4a8cf180-ab47-11ea-974b-de3912d90c2d.gif" width="350" height="250"> 

### Markov Soccer
&nbsp; &nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;<img src="https://user-images.githubusercontent.com/37489792/84292900-4bbe1e80-ab47-11ea-9c05-a0fd09eb346d.gif" width="350" height="250">&nbsp; &nbsp;&nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp;&nbsp;&nbsp; <img src="https://user-images.githubusercontent.com/37489792/84292899-4b258800-ab47-11ea-9e2e-27f9ba5bdecb.gif" width="350" height="250"> 

### Matching Pennies
&nbsp; &nbsp;&nbsp;&nbsp;&nbsp; &nbsp;&nbsp;<img src="https://user-images.githubusercontent.com/37489792/84292901-4c56b500-ab47-11ea-9ba2-02d898240e5d.gif" width="350" height="250">&nbsp; &nbsp;&nbsp;&nbsp; <img src="https://user-images.githubusercontent.com/37489792/84293659-6349d700-ab48-11ea-86a7-29957b828584.gif" width="70" height="250">&nbsp; &nbsp;&nbsp; &nbsp; <img src="https://user-images.githubusercontent.com/37489792/84292903-4cef4b80-ab47-11ea-820c-eea200f85a7a.gif" width="350" height="250"> 

The gifs of iterations shown here might not align exactly with gifs of CoPG and GDA in MP and RPS game. Because it is always running on a anonymous git server. This will be fixed by using gif link from repositiory.  
